/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: imaging_fdata.c,v 1.10 2007/11/23 13:07:59 jaffe Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: imaging_fdata.c,v 1.10 2007/11/23 13:07:59 jaffe Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*----------------------------------------------------------------------- */
/* Imaging_data_table (float version)  class for FITS binary tables.      */
/*----------------------------------------------------------------------- */
#include "imaging_fdata.h"
#include "imaging_detector.h"

/* internal prototypes */
/* get data from table_def specific to imaging_fdata table */
void get_imaging_fdata_info (imaging_fdata_table_def *table);

/* file global variables for table type*/
static char *tabletype = "IMAGING_DATA";


/* imaging_fdata_row functions */
imaging_fdata_row* make_imaging_fdata_row(imaging_fdata_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an imaging_fdata_row structure                                  */
/*   Inputs:                                                              */
/*      table     pointer to imaging_fdata table structure                */
/*   Returns:                                                             */
/*      An imaging_fdata_row structure pointer ; NULL indicates failure   */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  imaging_fdata_row *out;
  int i, icol, nregion;

  /* create basic structure */
  out = (imaging_fdata_row*)malloc(sizeof(imaging_fdata_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(imaging_fdata_row));
  
  nregion = table->nregion;
  out->nregion = table->nregion;

  /* allocate arrays */
  out->opt_train = (int*)malloc(table->detector->maxtel*sizeof(int));
  memset (out->opt_train, 0, table->detector->maxtel*sizeof(int));
  out->targtype = (char*)malloc(nregion*sizeof(char));
  memset (out->targtype, 'U', nregion*sizeof(char));
  out->target = (int*)malloc(nregion*sizeof(int));
  memset (out->target, 0, nregion*sizeof(int));
  out->data     = (fDataType**)malloc(nregion*sizeof(float*));

  /* Make generic structure */
  out->raw = make_table_row_from_def(table->def);

  /* attach data pointers to those in generic structure*/
  /* When the generic row is read/written the generic array is */
  /* automatically updated */
  for (i=0; i<table->nregion; i++) {
    icol = table->data_col[i];  /* which column? */
    out->data[i] = (fDataType*)out->raw->data[icol]->tdata.tvoid; 
  }

  return out; /* done */
} /* end  make_imaging_fdata_row */

void kill_imaging_fdata_row (imaging_fdata_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an imaging_fdata_row structure                                */
/*   Inputs:                                                              */
/*      row    pointer to the imaging_fdata_row structure.                */
/*----------------------------------------------------------------------- */
{
  int i, nregion;
  if (row==NULL) return; /* sanity check */
  if (row->opt_train!=NULL) free(row->opt_train);
  if (row->targtype!=NULL) free(row->targtype);
  if (row->data!=NULL) {
    for (i=0; i<row->nregion; i++) 
      if (row->data[i]!=NULL) free(row->data[i]);
    free(row->data);
  }
  free (row->raw); /* generic structure */
  free (row);      /* basic structure */
} /* end  kill_imaging_fdata_row */


/* imaging_fdata_table functions */
imaging_fdata_table_def* 
make_imaging_fdata_table_def(fitsfile *fptr, 
			    imaging_detector_table_def *detector,
			    int maxstep, int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new imaging_fdata_table in an open FITS file                  */
/*   Writes imaging detector table to FITS file;                          */
/*   then opens imaging data table                                        */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*                NULL -> don't write anything to file                    */
/*      detector  imaging_detector_table describing table                 */
/*      maxstep   Maximum number of steps in obs. sequence.               */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An imaging_fdata_table_def pointer; NULL indicates failure        */
/*----------------------------------------------------------------------- */
{
  imaging_fdata_table_def *out = NULL;
  int                    version = 1; /* table version number */
  int                    ikey, icol, i;
  long                   nx=0, ny=0, nregion;

  if (*ierr>0) return NULL;    /* existing error condition? */
/* fptr = NULL means create table def without connecting to a file
 * this is useful if you want to use ouput in make_imaging_fdata_row */
/*  if (fptr==NULL)     return NULL;*/ 
  if (detector==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (imaging_fdata_table_def*)malloc(sizeof(imaging_fdata_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(imaging_fdata_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(IMAGING_FDATA_TABLE, 'n', detector->nregion);
  if (out->def==NULL) { /* check for failure */
    kill_imaging_fdata_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating IMAGING_FDATA_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* attach imaging_detector_table */
  out->detector = detector;

  /* add keywords */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);
  ikey = table_find_key(out->def, "NREGION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&detector->nregion, ierr);
  ikey = table_find_key(out->def, "MAXTEL");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&detector->maxtel, ierr);
  ikey = table_find_key(out->def, "MAXSTEP"); /* default = 1*/
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&maxstep, ierr);
  ikey = table_find_key(out->def, "DATE-OBS");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)detector->date_obs, ierr);

  /* copy info from table descriptor */
  get_imaging_fdata_info (out);

  /* set size of opt_train member */
  if (out->opt_tran_col>=0) 
    out->def->colcount[out->opt_tran_col] = detector->maxtel;

  /* loop over regions setting sizes, dimensions */
  nregion = out->detector->nregion;  /* how many regions */
  for (i=0; i<nregion; i++) {
    icol = out->data_col[i]; 
    /* set data array size and dimension */
    /* get size from Imaging Detector structure */
    nx = detector->naxis[i][0];
    ny = detector->naxis[i][1];
    /* data array size */
    out->def->colcount[icol] =  nx*ny;
    /* specify FITS type code */
    out->def->coldatatype[icol] = FDATACODE;
    /* dimensionality string */
    if (out->def->coldim[icol]) free (out->def->coldim[icol]); /* free if it exists */
    out->def->coldim[icol] = (char*)malloc(FLEN_VALUE+1);      /* allocate */
    sprintf (out->def->coldim[icol],"(%ld,%ld)",nx,ny);        /* set dimensionality */
  } /* end loop over regions */

  /* write imaging_detector_table */
  if (fptr != NULL) write_imaging_detector_table(fptr, detector, ierr);

  /* open imaging data table */
  if (fptr != NULL) make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_imaging_fdata_table */

imaging_fdata_table_def* open_imaging_fdata_table(fitsfile *fptr,
						int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing imaging_fdata_table in the active FITS file         */
/*   Reads imaging_detector_table and saves in imaging_fdata_table_def    */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An imaging_fdata_table_def pointer; NULL indicates failure        */
/*----------------------------------------------------------------------- */
{
  imaging_fdata_table_def *out = NULL;
  int ikey;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (imaging_fdata_table_def*)malloc(sizeof(imaging_fdata_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(imaging_fdata_table_def)); /* zero fill */

  /* read imaging_detector_table */
  out->detector = read_imaging_detector_table(fptr, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR reading IMAGING_FDATA_TABLE for %s table_def \n", 
	    tabletype);
    kill_imaging_fdata_table_def(out); /* clean up */
    return NULL;   /* bomb out */
  }

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_imaging_fdata_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_imaging_fdata_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* keywords */
  out->maxstep = 1;  /* MAXSTEP */
  ikey = table_find_key(out->def, "MAXSTEP");
  if (ikey>0) table_get_keyword(out->def, ikey, (void*)&out->maxstep, ierr);

  /* copy info from table descriptor */
  get_imaging_fdata_info (out);

  return out; /* done */

} /* end open_imaging_fdata_table */

int read_imaging_fdata_table(fitsfile *fptr, imaging_fdata_table_def* table, 
			    imaging_fdata_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an imaging_fdata_table                        */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_imaging_fdata_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  row->time        = trow->data[table->time_col]->tdata.tdouble[0];
  if (table->def->coldim[table->frame_col]!=NULL)
    row->frame     = trow->data[table->frame_col]->tdata.tlong[0];
  row->opd[0]         = trow->data[table->opd_col]->tdata.tdouble[0];
  row->opd[1]         = trow->data[table->opd_col]->tdata.tdouble[1];
  row->local_opd[0]   = trow->data[table->local_opd_col]->tdata.tdouble[0];
  row->local_opd[1]   = trow->data[table->local_opd_col]->tdata.tdouble[1];
  row->exp         = trow->data[table->exp_col]->tdata.tfloat[0];
  row->offset[0]   = trow->data[table->offset_col]->tdata.tfloat[0];
  row->offset[1]   = trow->data[table->offset_col]->tdata.tfloat[1];
  row->rotation    = trow->data[table->rotation_col]->tdata.tfloat[0];
  if (table->def->coldim[table->reference_col]!=NULL) /* only if there */
    row->reference   = trow->data[table->reference_col]->tdata.tshort[0];
  row->step        = trow->data[table->step_col]->tdata.tshort[0];
  for (i=0; i<table->detector->maxtel; i++) {
    row->opt_train[i] = trow->data[table->opt_tran_col]->tdata.tshort[i];
  }
  for (i=0; i<table->nregion; i++) {
    row->target[i] = trow->data[table->target_col[i]]->tdata.tshort[0];
    row->targtype[i] = trow->data[table->targtype_col[i]]->tdata.tchar[0];
  }

  return 0; /* OK - done */
} /* end read_imaging_fdata_table */

int write_imaging_fdata_table(fitsfile *fptr, imaging_fdata_table_def* table, 
			     imaging_fdata_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an imaging_fdata_table                        */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */     
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  trow->data[table->time_col]->tdata.tdouble[0]      = row->time;
  if (table->def->coldim[table->frame_col]!=NULL)
    trow->data[table->frame_col]->tdata.tlong[0]     = row->frame;
  trow->data[table->opd_col]->tdata.tdouble[0]       = row->opd[0];
  trow->data[table->opd_col]->tdata.tdouble[1]       = row->opd[1];
  trow->data[table->local_opd_col]->tdata.tdouble[0] = row->local_opd[0];
  trow->data[table->local_opd_col]->tdata.tdouble[1] = row->local_opd[1];
  trow->data[table->exp_col]->tdata.tfloat[0]        = row->exp;
  trow->data[table->offset_col]->tdata.tfloat[0]     = row->offset[0];
  trow->data[table->offset_col]->tdata.tfloat[1]     = row->offset[1];
  trow->data[table->rotation_col]->tdata.tfloat[0]   = row->rotation;
  if (table->def->coldim[table->reference_col]!=NULL)   /* only if there */
    trow->data[table->reference_col]->tdata.tshort[0]= row->reference;
  trow->data[table->step_col]->tdata.tshort[0]       = row->step;
  for (i=0; i<table->detector->maxtel; i++) {
    trow->data[table->opt_tran_col]->tdata.tshort[i] = 
      row->opt_train[i];
  }
  for (i=0; i<table->nregion; i++) {
    char temp[2] = {'\0','\0'};
    temp[0] = row->targtype[i];
    strncpy (trow->data[table->targtype_col[i]]->tdata.tchar,temp,2);
    trow->data[table->target_col[i]]->tdata.tshort[0]  = row->target[i];
  }

  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_imaging_fdata_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_imaging_fdata_table */

int close_imaging_fdata_table(fitsfile *fptr, imaging_fdata_table_def* table, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open imaging_fdata_table, associated structures are destroyed*/
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_imaging_fdata_table_def(table);

  return 0; /* OK - done */
} /* end close_imaging_fdata_table */

void kill_imaging_fdata_table_def (imaging_fdata_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an imaging_fdata_table structure                              */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  if (table->def!=NULL) 
		kill_table_def(table->def);
  /* detector table definition structure */
  if (table->detector!=NULL) 
		kill_imaging_detector_table_def(table->detector);

  /* free keyword arrays */
  /* date_obs is pointer to detector value */

  /* free arrays */
  if (table->targtype_col!=NULL) free (table->targtype_col);
  if (table->data_col!=NULL) free (table->data_col);

  /* free basic structure */
  free (table);
} /* end kill_imaging_fdata_table_def */

/*************************private functions ******************************/

void get_imaging_fdata_info (imaging_fdata_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to imaging_fdata table              */
/*   Uses detector member which must already be attached                  */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  char  tstring[81];
  int   i, nregion;

  if (table==NULL) return;             /* sanity check */
  if (table->detector==NULL) return;   /* sanity check */

  /* keyword values */
  /* simply use pointer for date_obs value in detector */
  table->date_obs = table->detector->date_obs;

  /* lookup column numbers (0-rel) */
  table->time_col      = table_find_col (table->def, "TIME")-1;
  table->frame_col     = table_find_col (table->def, "FRAME")-1;
  table->exp_col       = table_find_col (table->def, "EXPTIME")-1;
  table->opd_col       = table_find_col (table->def, "OPD")-1;
  table->opt_tran_col  = table_find_col (table->def, "OPT_TRAIN")-1;
  table->ins_train_col  = table_find_col (table->def, "INS_TRAIN")-1;
  table->local_opd_col = table_find_col (table->def, "LOCALOPD")-1;
  table->offset_col    = table_find_col (table->def, "OFFSET")-1;
  table->rotation_col  = table_find_col (table->def, "ROTATION")-1;
  table->reference_col = table_find_col (table->def, "REFERENCE")-1;
  table->step_col      = table_find_col (table->def, "STEPPING_PHASE")-1;

  /* Some column types are one per region */
  nregion = table->detector->nregion;  /* how many regions */
  table->nregion = nregion;
  /* allocate arrays */
  table->target_col   = (int*)malloc(nregion*sizeof(int*));
  table->targtype_col = (int*)malloc(nregion*sizeof(int*));
  table->data_col     = (int*)malloc(nregion*sizeof(int*));
  /* loop over regions finding column numbers */
  for (i=0; i<nregion; i++) {
    sprintf (tstring, "TARGET%d",i+1);
    table->target_col[i] = table_find_col (table->def, tstring)-1;
    sprintf (tstring, "TARTYP%d",i+1);
    table->targtype_col[i] = table_find_col (table->def, tstring)-1;
    sprintf (tstring, "DATA%d",i+1);
    table->data_col[i] = table_find_col (table->def, tstring)-1;
  } /* end loop over regions */
} /* end  get_imaging_fdata_info */

/* */
